from typing import List

from beanie import Document, Link, Indexed


class WindowAPI(Document):
    x: int
    y: int


class DoorAPI(Document):
    t: int = 10


class RoofAPI(Document):
    r: int = 100


class HouseAPI(Document):
    windows: List[Link[WindowAPI]]
    name: Indexed(str)
    height: Indexed(int) = 2



from typing import Optional, Union

from beanie.odm.utils.typing import extract_id_class


class TestTyping:
    def test_extract_id_class(self):
        # Union
        assert extract_id_class(Union[str, int]) == str
        assert extract_id_class(Union[str, None]) == str
        assert extract_id_class(Union[str, None, int]) == str
        # Optional
        assert extract_id_class(Optional[str]) == str



import asyncio
from typing import Any, ClassVar, Dict, Optional, Union
from beanie.odm.fields import Link, LinkInfo

from pydantic import BaseModel

from beanie.exceptions import ViewWasNotInitialized
from beanie.odm.interfaces.aggregate import AggregateInterface
from beanie.odm.interfaces.detector import DetectionInterface, ModelType
from beanie.odm.interfaces.find import FindInterface
from beanie.odm.interfaces.getters import OtherGettersInterface
from beanie.odm.settings.view import ViewSettings


class View(
    BaseModel,
    FindInterface,
    AggregateInterface,
    OtherGettersInterface,
    DetectionInterface,
):
    """
    What is needed:

    Source collection or view
    pipeline

    """

    # Relations
    _link_fields: ClassVar[Optional[Dict[str, LinkInfo]]] = None

    # Settings
    _settings: ClassVar[ViewSettings]

    @classmethod
    def get_settings(cls) -> ViewSettings:
        """
        Get view settings, which was created on
        the initialization step

        :return: ViewSettings class
        """
        if cls._settings is None:
            raise ViewWasNotInitialized
        return cls._settings

    async def fetch_link(self, field: Union[str, Any]):
        ref_obj = getattr(self, field, None)
        if isinstance(ref_obj, Link):
            value = await ref_obj.fetch(fetch_links=True)
            setattr(self, field, value)
        if isinstance(ref_obj, list) and ref_obj:
            values = await Link.fetch_list(ref_obj, fetch_links=True)
            setattr(self, field, values)

    async def fetch_all_links(self):
        coros = []
        link_fields = self.get_link_fields()
        if link_fields is not None:
            for ref in link_fields.values():
                coros.append(self.fetch_link(ref.field_name))  # TODO lists
        await asyncio.gather(*coros)

    @classmethod
    def get_link_fields(cls) -> Optional[Dict[str, LinkInfo]]:
        return cls._link_fields

    @classmethod
    def get_model_type(cls) -> ModelType:
        return ModelType.View



from typing import Type, Any

import pydantic
from pydantic import BaseModel

IS_PYDANTIC_V2 = int(pydantic.VERSION.split(".")[0]) >= 2

if IS_PYDANTIC_V2:
    from pydantic import TypeAdapter
else:
    from pydantic import parse_obj_as


def parse_object_as(object_type: Type, data: Any):
    if IS_PYDANTIC_V2:
        return TypeAdapter(object_type).validate_python(data)
    else:
        return parse_obj_as(object_type, data)


def get_field_type(field):
    if IS_PYDANTIC_V2:
        return field.annotation
    else:
        return field.type_


def get_model_fields(model):
    if IS_PYDANTIC_V2:
        return model.model_fields
    else:
        return model.__fields__


def parse_model(model_type: Type[BaseModel], data: Any):
    if IS_PYDANTIC_V2:
        return model_type.model_validate(data)
    else:
        return model_type.parse_obj(data)


def get_extra_field_info(field, parameter: str):
    if IS_PYDANTIC_V2:
        if field.json_schema_extra is not None:
            return field.json_schema_extra.get(parameter)
        return None
    else:
        return field.field_info.extra.get(parameter)


def get_config_value(model, parameter: str):
    if IS_PYDANTIC_V2:
        return model.model_config.get(parameter)
    else:
        return getattr(model.Config, parameter, None)


def get_model_dump(model):
    if IS_PYDANTIC_V2:
        return model.model_dump()
    else:
        return model.dict()



from tests.odm.models import DocumentTestModel


async def test_count_with_filter_query(documents):
    await documents(4, "uno", True)
    await documents(2, "dos", True)
    await documents(1, "cuatro", True)
    e = await DocumentTestModel.find_many({"test_str": "dos"}).exists()
    assert e is True

    e = await DocumentTestModel.find_one({"test_str": "dos"}).exists()
    assert e is True

    e = await DocumentTestModel.find_many({"test_str": "wrong"}).exists()
    assert e is False



from beanie.odm.operators.find.array import All, ElemMatch, Size
from tests.odm.models import Sample, PackageElemMatch


async def test_all():
    q = All(Sample.integer, [1, 2, 3])
    assert q == {"integer": {"$all": [1, 2, 3]}}


async def test_elem_match():
    q = ElemMatch(Sample.integer, {"a": "b"})
    assert q == {"integer": {"$elemMatch": {"a": "b"}}}


async def test_size():
    q = Size(Sample.integer, 4)
    assert q == {"integer": {"$size": 4}}


async def test_elem_match_nested():
    q = ElemMatch(
        PackageElemMatch.releases, major_ver=7, minor_ver=1, build_ver=0
    )
    assert q == {
        "releases": {
            "$elemMatch": {"major_ver": 7, "minor_ver": 1, "build_ver": 0}
        }
    }



from datetime import datetime, timedelta
from random import randint
from typing import List

import pytest

from beanie.odm.utils.init import init_beanie
from tests.odm.models import (
    Bicycle,
    Bike,
    Bus,
    Car,
    DocumentForEncodingTest,
    DocumentForEncodingTestDate,
    DocumentWithStringField,
    DocumentMultiModelOne,
    DocumentMultiModelTwo,
    DocumentTestModel,
    DocumentTestModelWithLink,
    DocumentTestModelFailInspection,
    DocumentTestModelWithComplexIndex,
    DocumentTestModelWithCustomCollectionName,
    DocumentTestModelWithIndexFlags,
    DocumentTestModelWithIndexFlagsAliases,
    DocumentTestModelWithSimpleIndex,
    DocumentUnion,
    DocumentWithActions,
    DocumentWithActions2,
    DocumentWithBsonEncodersFiledsTypes,
    DocumentWithCustomFiledsTypes,
    DocumentWithCustomIdInt,
    DocumentWithCustomIdUUID,
    DocumentWithExtras,
    DocumentWithPydanticConfig,
    DocumentWithRevisionTurnedOn,
    DocumentWithTurnedOffStateManagement,
    DocumentWithTurnedOnReplaceObjects,
    DocumentWithTurnedOnStateManagement,
    DocumentWithTurnedOnSavePrevious,
    DocumentWithValidationOnSave,
    Door,
    GeoObject,
    House,
    HouseWithRevision,
    InheritedDocumentWithActions,
    Lock,
    LockWithRevision,
    Nested,
    Option1,
    Option2,
    Owner,
    Roof,
    Sample,
    SampleLazyParsing,
    SampleWithMutableObjects,
    SubDocument,
    Test2NonRoot,
    TestNonRoot,
    Vehicle,
    Window,
    WindowWithRevision,
    Yard,
    YardWithRevision,
    RootDocument,
    ADocument,
    BDocument,
    StateAndDecimalFieldModel,
    Region,
    UsersAddresses,
    SelfLinked,
    LoopedLinksA,
    LoopedLinksB,
    DocumentWithTurnedOnStateManagementWithCustomId,
    DocumentWithDecimalField,
    DocumentWithKeepNullsFalse,
    PackageElemMatch,
    DocumentWithLink,
    DocumentWithBackLink,
    DocumentWithListBackLink,
    DocumentWithListLink,
    DocumentWithListOfLinks,
    DocumentToBeLinked,
    DocumentWithTimeStampToTestConsistency,
    DocumentWithIndexMerging1,
    DocumentWithIndexMerging2,
    DocumentWithCustomInit,
)
from tests.odm.views import TestView, TestViewWithLink


@pytest.fixture
def point():
    return {
        "longitude": 13.404954,
        "latitude": 52.520008,
    }


@pytest.fixture
async def preset_documents(point):
    docs = []
    for i in range(10):
        timestamp = datetime.utcnow() - timedelta(days=i)
        integer_1: int = i // 3
        integer_2: int = i // 2
        float_num = integer_1 + 0.3
        string: str = f"test_{integer_1}"
        option_1 = Option1(s="TEST")
        option_2 = Option2(f=3.14)
        union = option_1 if i % 2 else option_2
        optional = option_2 if not i % 3 else None
        geo = GeoObject(
            coordinates=[
                point["longitude"] + i / 10,
                point["latitude"] + i / 10,
            ]
        )
        nested = Nested(
            integer=integer_2,
            option_1=option_1,
            union=union,
            optional=optional,
        )

        const = "TEST"

        sample = Sample(
            timestamp=timestamp,
            increment=i,
            integer=integer_1,
            float_num=float_num,
            string=string,
            nested=nested,
            optional=optional,
            union=union,
            geo=geo,
            const=const,
        )

        docs.append(sample)
    await Sample.insert_many(documents=docs)


@pytest.fixture()
def sample_doc_not_saved(point):
    nested = Nested(
        integer=0,
        option_1=Option1(s="TEST"),
        union=Option1(s="TEST"),
        optional=None,
    )
    geo = GeoObject(
        coordinates=[
            point["longitude"],
            point["latitude"],
        ]
    )
    return Sample(
        timestamp=datetime.utcnow(),
        increment=0,
        integer=0,
        float_num=0,
        string="TEST_NOT_SAVED",
        nested=nested,
        optional=None,
        union=Option1(s="TEST"),
        geo=geo,
    )


@pytest.fixture()
async def session(cli):
    s = await cli.start_session()
    yield s
    await s.end_session()


@pytest.fixture(autouse=True)
async def init(db):
    models = [
        DocumentWithExtras,
        DocumentWithPydanticConfig,
        DocumentTestModel,
        DocumentTestModelWithLink,
        DocumentTestModelWithCustomCollectionName,
        DocumentTestModelWithSimpleIndex,
        DocumentTestModelWithIndexFlags,
        DocumentTestModelWithIndexFlagsAliases,
        DocumentTestModelWithComplexIndex,
        DocumentTestModelFailInspection,
        DocumentWithBsonEncodersFiledsTypes,
        DocumentWithCustomFiledsTypes,
        DocumentWithCustomIdUUID,
        DocumentWithCustomIdInt,
        Sample,
        DocumentWithActions,
        DocumentWithTurnedOnStateManagement,
        DocumentWithTurnedOnReplaceObjects,
        DocumentWithTurnedOnSavePrevious,
        DocumentWithTurnedOffStateManagement,
        DocumentWithValidationOnSave,
        DocumentWithRevisionTurnedOn,
        House,
        Window,
        Door,
        Roof,
        Yard,
        Lock,
        InheritedDocumentWithActions,
        DocumentForEncodingTest,
        DocumentForEncodingTestDate,
        DocumentWithStringField,
        TestView,
        TestViewWithLink,
        DocumentMultiModelOne,
        DocumentMultiModelTwo,
        DocumentUnion,
        HouseWithRevision,
        WindowWithRevision,
        LockWithRevision,
        YardWithRevision,
        DocumentWithActions2,
        Vehicle,
        Bicycle,
        Bike,
        Car,
        Bus,
        Owner,
        SampleWithMutableObjects,
        TestNonRoot,
        Test2NonRoot,
        SampleLazyParsing,
        RootDocument,
        ADocument,
        BDocument,
        StateAndDecimalFieldModel,
        Region,
        UsersAddresses,
        SelfLinked,
        LoopedLinksA,
        LoopedLinksB,
        DocumentWithTurnedOnStateManagementWithCustomId,
        DocumentWithDecimalField,
        DocumentWithKeepNullsFalse,
        PackageElemMatch,
        DocumentWithLink,
        DocumentWithBackLink,
        DocumentWithListLink,
        DocumentWithListBackLink,
        DocumentWithListOfLinks,
        DocumentToBeLinked,
        DocumentWithTimeStampToTestConsistency,
        DocumentWithIndexMerging1,
        DocumentWithIndexMerging2,
        DocumentWithCustomInit,
    ]
    await init_beanie(
        database=db,
        document_models=models,
    )

    yield None

    for model in models:
        await model.get_motor_collection().drop()
        await model.get_motor_collection().drop_indexes()


@pytest.fixture
def document_not_inserted():
    return DocumentTestModel(
        test_int=42,
        test_list=[SubDocument(test_str="foo"), SubDocument(test_str="bar")],
        test_doc=SubDocument(test_str="foobar"),
        test_str="kipasa",
    )


@pytest.fixture
def documents_not_inserted():
    def generate_documents(
        number: int, test_str: str = None, random: bool = False
    ) -> List[DocumentTestModel]:
        return [
            DocumentTestModel(
                test_int=randint(0, 1000000) if random else i,
                test_list=[
                    SubDocument(test_str="foo"),
                    SubDocument(test_str="bar"),
                ],
                test_doc=SubDocument(test_str="foobar"),
                test_str="kipasa" if test_str is None else test_str,
            )
            for i in range(number)
        ]

    return generate_documents


@pytest.fixture
async def document(document_not_inserted) -> DocumentTestModel:
    return await document_not_inserted.insert()


@pytest.fixture
def documents(documents_not_inserted):
    async def generate_documents(
        number: int, test_str: str = None, random: bool = False
    ):
        result = await DocumentTestModel.insert_many(
            documents_not_inserted(number, test_str, random)
        )
        return result.inserted_ids

    return generate_documents


@pytest.fixture
def documents_with_links(documents):
    async def generate_documents():
        await documents(15)
        results = await DocumentTestModel.all().to_list()
        for document in results:
            await DocumentTestModelWithLink(test_link=document).insert()

    return generate_documents



import pytest
from bson import ObjectId

from beanie import PydanticObjectId, WriteRules
from beanie.exceptions import StateManagementIsTurnedOff, StateNotSaved
from beanie.odm.utils.parsing import parse_obj
from beanie.odm.utils.pydantic import IS_PYDANTIC_V2, parse_model
from tests.odm.models import (
    DocumentWithTurnedOffStateManagement,
    DocumentWithTurnedOnReplaceObjects,
    DocumentWithTurnedOnStateManagement,
    DocumentWithTurnedOnSavePrevious,
    HouseWithRevision,
    InternalDoc,
    LockWithRevision,
    WindowWithRevision,
    StateAndDecimalFieldModel,
    DocumentWithTurnedOnStateManagementWithCustomId,
)


@pytest.fixture
def state():
    if IS_PYDANTIC_V2:
        internal = InternalDoc().model_dump()
    else:
        internal = InternalDoc().dict()
    return {
        "num_1": 1,
        "num_2": 2,
        "_id": ObjectId(),
        "internal": internal,
    }


@pytest.fixture
def state_without_id():
    if IS_PYDANTIC_V2:
        internal = InternalDoc().model_dump()
    else:
        internal = InternalDoc().dict()

    return {
        "num_1": 1,
        "num_2": 2,
        "internal": internal,
    }


@pytest.fixture
def doc_default(state):
    return parse_obj(DocumentWithTurnedOnStateManagement, state)


@pytest.fixture
def doc_replace(state):
    return parse_obj(DocumentWithTurnedOnReplaceObjects, state)


@pytest.fixture
def doc_previous(state):
    return parse_obj(DocumentWithTurnedOnSavePrevious, state)


@pytest.fixture
async def saved_doc_default(doc_default):
    await doc_default.insert()
    return doc_default


@pytest.fixture
async def saved_doc_previous(doc_previous):
    await doc_previous.insert()
    return doc_previous


@pytest.fixture
def windows_not_inserted():
    return [
        WindowWithRevision(x=10, y=10, lock=LockWithRevision(k=10)),
        WindowWithRevision(x=11, y=11, lock=LockWithRevision(k=11)),
    ]


@pytest.fixture
def house_not_inserted(windows_not_inserted):
    return HouseWithRevision(windows=windows_not_inserted)


@pytest.fixture
async def house(house_not_inserted):
    return await house_not_inserted.insert(link_rule=WriteRules.WRITE)


class TestStateManagement:
    async def test_use_state_management_property(self):
        assert (
            DocumentWithTurnedOnStateManagement.use_state_management() is True
        )
        assert (
            DocumentWithTurnedOffStateManagement.use_state_management()
            is False
        )

    async def test_state_with_decimal_field(
        self,
    ):
        await StateAndDecimalFieldModel(amt=10.01).insert()
        await StateAndDecimalFieldModel.all().to_list()

    async def test_parse_object_with_saving_state(self):
        if IS_PYDANTIC_V2:
            internal = InternalDoc().model_dump()
        else:
            internal = InternalDoc().dict()
        obj = {
            "num_1": 1,
            "num_2": 2,
            "_id": ObjectId(),
            "internal": internal,
        }
        doc = parse_obj(DocumentWithTurnedOnStateManagement, obj)
        assert doc.get_saved_state() == obj
        assert doc.get_previous_saved_state() is None

    class TestSaveState:
        async def test_save_state(self):
            doc = DocumentWithTurnedOnStateManagement(
                num_1=1, num_2=2, internal=InternalDoc(num=1, string="s")
            )
            assert doc.get_saved_state() is None
            assert doc.get_previous_saved_state() is None

            doc.id = PydanticObjectId()
            doc._save_state()
            assert doc.get_saved_state() == {
                "num_1": 1,
                "num_2": 2,
                "internal": {"num": 1, "string": "s", "lst": [1, 2, 3, 4, 5]},
                "_id": doc.id,
            }
            assert doc.get_previous_saved_state() is None

            doc.num_1 = 2
            doc.num_2 = 3
            doc._save_state()
            assert doc.get_saved_state() == {
                "num_1": 2,
                "num_2": 3,
                "internal": {"num": 1, "string": "s", "lst": [1, 2, 3, 4, 5]},
                "_id": doc.id,
            }
            assert doc.get_previous_saved_state() is None

        async def test_save_state_with_custom_id_type(self):
            doc = DocumentWithTurnedOnStateManagementWithCustomId(
                id=0,
                num_1=1,
                num_2=2,
            )
            with pytest.raises(StateNotSaved):
                await doc.save_changes()
            doc.num_1 = 2
            with pytest.raises(StateNotSaved):
                await doc.save_changes()

        async def test_save_state_with_previous(self):
            doc = DocumentWithTurnedOnSavePrevious(
                num_1=1, num_2=2, internal=InternalDoc(num=1, string="s")
            )
            assert doc.get_saved_state() is None
            assert doc.get_previous_saved_state() is None

            doc.id = PydanticObjectId()
            doc._save_state()
            assert doc.get_saved_state() == {
                "num_1": 1,
                "num_2": 2,
                "internal": {"num": 1, "string": "s", "lst": [1, 2, 3, 4, 5]},
                "_id": doc.id,
            }
            assert doc.get_previous_saved_state() is None

            doc.num_1 = 2
            doc.num_2 = 3
            doc._save_state()
            assert doc.get_saved_state() == {
                "num_1": 2,
                "num_2": 3,
                "internal": {"num": 1, "string": "s", "lst": [1, 2, 3, 4, 5]},
                "_id": doc.id,
            }
            assert doc.get_previous_saved_state() == {
                "num_1": 1,
                "num_2": 2,
                "internal": {"num": 1, "string": "s", "lst": [1, 2, 3, 4, 5]},
                "_id": doc.id,
            }

    class TestIsChanged:
        async def test_state_management_off(self):
            doc = DocumentWithTurnedOffStateManagement(num_1=1, num_2=2)

            with pytest.raises(StateManagementIsTurnedOff):
                doc.is_changed

        async def test_state_management_on_not_changed(self):
            doc = DocumentWithTurnedOnStateManagement(
                num_1=1, num_2=2, internal=InternalDoc()
            )

            with pytest.raises(StateNotSaved):
                doc.is_changed

        async def test_state_management_on_changed(self, doc_default):
            assert doc_default.is_changed is False

            doc_default.num_1 = 10

            assert doc_default.is_changed is True

    class TestHasChanged:
        async def test_state_management_off(self):
            doc = DocumentWithTurnedOffStateManagement(num_1=1, num_2=2)

            with pytest.raises(StateManagementIsTurnedOff):
                doc.has_changed

        async def test_state_management_on_not_changed(self):
            doc = DocumentWithTurnedOnStateManagement(
                num_1=1, num_2=2, internal=InternalDoc()
            )

            with pytest.raises(StateNotSaved):
                doc.has_changed

        async def test_save_previous_on_not_changed(self):
            doc = DocumentWithTurnedOnSavePrevious(
                num_1=1, num_2=2, internal=InternalDoc()
            )

            with pytest.raises(StateNotSaved):
                doc.has_changed

        async def test_save_previous_on_changed(self, doc_previous):
            assert doc_previous.has_changed is False

            doc_previous.num_1 = 10
            doc_previous._save_state()

            assert doc_previous.has_changed is True

    class TestGetChanges:
        async def test_valid(self, doc_default):
            doc_default.internal.num = 1000
            doc_default.internal.string = "new_value"
            doc_default.internal.lst.append(100)

            assert doc_default.get_changes() == {
                "internal.num": 1000,
                "internal.string": "new_value",
                "internal.lst": [1, 2, 3, 4, 5, 100],
            }

            doc_default._save_state()

            assert doc_default.get_changes() == {}

        async def test_whole(self, doc_default):
            doc_default.internal = {"num": 1000, "string": "new_value"}

            assert doc_default.get_changes() == {
                "internal.num": 1000,
                "internal.string": "new_value",
            }

        async def test_replace(self, doc_replace):
            doc_replace.internal.num = 1000
            doc_replace.internal.string = "new_value"

            assert doc_replace.get_changes() == {
                "internal": {
                    "num": 1000,
                    "string": "new_value",
                    "lst": [1, 2, 3, 4, 5],
                }
            }

        async def test_replace_whole(self, doc_replace):
            doc_replace.internal = {"num": 1000, "string": "new_value"}

            assert doc_replace.get_changes() == {
                "internal": {
                    "num": 1000,
                    "string": "new_value",
                }
            }

    class TestGetPreviousChanges:
        async def test_get_previous_changes(self, doc_previous):
            doc_previous.internal.num = 1000
            doc_previous.internal.string = "new_value"
            doc_previous.internal.lst.append(100)

            assert doc_previous.get_previous_changes() == {}

            doc_previous._save_state()

            assert doc_previous.get_previous_changes() == {
                "internal.num": 1000,
                "internal.string": "new_value",
                "internal.lst": [1, 2, 3, 4, 5, 100],
            }

    class TestRollback:
        async def test_rollback(self, doc_default, state):
            doc_default.num_1 = 100
            doc_default.rollback()

            assert doc_default.num_1 == state["num_1"]

    class TestQueries:
        async def test_save_changes(self, saved_doc_default):
            assert saved_doc_default.get_saved_state()["num_1"] == 1
            assert saved_doc_default.get_previous_saved_state() is None

            saved_doc_default.num_1 = 10000

            saved_doc_default.internal.change_private()
            assert (
                saved_doc_default.internal.get_private() == "PRIVATE_CHANGED"
            )

            await saved_doc_default.save_changes()

            assert saved_doc_default.get_saved_state()["num_1"] == 10000
            assert saved_doc_default.get_previous_saved_state() is None
            assert (
                saved_doc_default.internal.get_private() == "PRIVATE_CHANGED"
            )

            new_doc = await DocumentWithTurnedOnStateManagement.get(
                saved_doc_default.id
            )
            assert new_doc.num_1 == 10000

        async def test_save_changes_previous(self, saved_doc_previous):
            assert saved_doc_previous.get_saved_state()["num_1"] == 1
            assert saved_doc_previous.get_previous_saved_state()["num_1"] == 1

            saved_doc_previous.num_1 = 10000

            saved_doc_previous.internal.change_private()
            assert (
                saved_doc_previous.internal.get_private() == "PRIVATE_CHANGED"
            )

            await saved_doc_previous.save_changes()
            assert saved_doc_previous.get_saved_state()["num_1"] == 10000
            assert saved_doc_previous.get_previous_saved_state()["num_1"] == 1
            assert (
                saved_doc_previous.internal.get_private() == "PRIVATE_CHANGED"
            )

            new_doc = await DocumentWithTurnedOnSavePrevious.get(
                saved_doc_previous.id
            )
            assert new_doc.num_1 == 10000

        async def test_fetch_save_changes(self, house):
            data = await HouseWithRevision.all(fetch_links=True).to_list()
            house = data[0]
            window_0 = house.windows[0]
            window_0.x = 10000
            window_0.lock.k = 10000
            await window_0.save_changes()

        async def test_find_one(self, saved_doc_default, state):
            new_doc = await DocumentWithTurnedOnStateManagement.get(
                saved_doc_default.id
            )
            assert new_doc.get_saved_state() == state
            assert new_doc.get_previous_saved_state() is None

            new_doc = await DocumentWithTurnedOnStateManagement.find_one(
                DocumentWithTurnedOnStateManagement.id == saved_doc_default.id
            )
            assert new_doc.get_saved_state() == state
            assert new_doc.get_previous_saved_state() is None

        async def test_find_many(self):
            docs = []
            for i in range(10):
                docs.append(
                    DocumentWithTurnedOnStateManagement(
                        num_1=i, num_2=i + 1, internal=InternalDoc()
                    )
                )
            await DocumentWithTurnedOnStateManagement.insert_many(docs)

            found_docs = await DocumentWithTurnedOnStateManagement.find(
                DocumentWithTurnedOnStateManagement.num_1 > 4
            ).to_list()

            for doc in found_docs:
                assert doc.get_saved_state() is not None
                assert doc.get_previous_saved_state() is None

        async def test_insert(self, state_without_id):
            doc = parse_model(
                DocumentWithTurnedOnStateManagement, state_without_id
            )
            assert doc.get_saved_state() is None
            await doc.insert()
            new_state = doc.get_saved_state()
            assert new_state["_id"] is not None
            del new_state["_id"]
            assert new_state == state_without_id

        async def test_replace(self, saved_doc_default):
            saved_doc_default.num_1 = 100
            await saved_doc_default.replace()

            assert saved_doc_default.get_saved_state()["num_1"] == 100
            assert saved_doc_default.get_previous_saved_state() is None

        async def test_replace_save_previous(self, saved_doc_previous):
            saved_doc_previous.num_1 = 100
            await saved_doc_previous.replace()

            assert saved_doc_previous.get_saved_state()["num_1"] == 100
            assert saved_doc_previous.get_previous_saved_state()["num_1"] == 1



from pydantic.main import BaseModel

from beanie import Document, free_fall_migration


class Tag(BaseModel):
    color: str
    name: str


class OldNote(Document):
    name: str
    tag: Tag

    class Settings:
        name = "notes"


class Note(Document):
    title: str
    tag: Tag

    class Settings:
        name = "notes"


class Forward:
    @free_fall_migration(document_models=[OldNote, Note])
    async def name_to_title(self, session):
        async for old_note in OldNote.find_all():
            new_note = Note(
                id=old_note.id, title=old_note.name, tag=old_note.tag
            )
            await new_note.replace(session=session)


class Backward:
    @free_fall_migration(document_models=[OldNote, Note])
    async def title_to_name(self, session):
        async for old_note in Note.find_all():
            new_note = OldNote(
                id=old_note.id, name=old_note.title, tag=old_note.tag
            )
            await new_note.replace(session=session)



from tests.odm.models import (
    DocumentMultiModelOne,
    DocumentMultiModelTwo,
    DocumentUnion,
)


class TestMultiModel:
    async def test_multi_model(self):
        doc_1 = await DocumentMultiModelOne().insert()
        doc_2 = await DocumentMultiModelTwo().insert()

        new_doc_1 = await DocumentMultiModelOne.get(doc_1.id)
        new_doc_2 = await DocumentMultiModelTwo.get(doc_2.id)

        assert new_doc_1 is not None
        assert new_doc_2 is not None

        new_doc_1 = await DocumentMultiModelTwo.get(doc_1.id)
        new_doc_2 = await DocumentMultiModelOne.get(doc_2.id)

        assert new_doc_1 is None
        assert new_doc_2 is None

        new_docs_1 = await DocumentMultiModelOne.find({}).to_list()
        new_docs_2 = await DocumentMultiModelTwo.find({}).to_list()

        assert len(new_docs_1) == 1
        assert len(new_docs_2) == 1

        await DocumentMultiModelOne.update_all({"$set": {"shared": 100}})

        new_doc_1 = await DocumentMultiModelOne.get(doc_1.id)
        new_doc_2 = await DocumentMultiModelTwo.get(doc_2.id)

        assert new_doc_1.shared == 100
        assert new_doc_2.shared == 0

    async def test_union_doc(self):
        await DocumentMultiModelOne().insert()
        await DocumentMultiModelTwo().insert()
        await DocumentMultiModelOne().insert()
        await DocumentMultiModelTwo().insert()

        docs = await DocumentUnion.all().to_list()
        assert isinstance(docs[0], DocumentMultiModelOne)
        assert isinstance(docs[1], DocumentMultiModelTwo)
        assert isinstance(docs[2], DocumentMultiModelOne)
        assert isinstance(docs[3], DocumentMultiModelTwo)

    async def test_union_doc_aggregation(self):
        await DocumentMultiModelOne().insert()
        await DocumentMultiModelTwo().insert()
        await DocumentMultiModelOne().insert()
        await DocumentMultiModelTwo().insert()

        docs = await DocumentUnion.aggregate(
            [{"$match": {"$expr": {"$eq": ["$int_filed", 0]}}}]
        ).to_list()
        assert len(docs) == 2

    async def test_union_doc_link(self):
        doc_1 = await DocumentMultiModelOne().insert()
        await DocumentMultiModelTwo(linked_doc=doc_1).insert()

        docs = await DocumentMultiModelTwo.find({}, fetch_links=True).to_list()
        assert isinstance(docs[0].linked_doc, DocumentMultiModelOne)



import pytest

from beanie import BulkWriter
from beanie.exceptions import RevisionIdWasChanged
from beanie.odm.operators.update.general import Inc
from tests.odm.models import DocumentWithRevisionTurnedOn
from pymongo.errors import BulkWriteError


async def test_replace():
    doc = DocumentWithRevisionTurnedOn(num_1=1, num_2=2)
    await doc.insert()

    doc.num_1 = 2
    await doc.replace()

    doc.num_2 = 3
    await doc.replace()

    for i in range(5):
        found_doc = await DocumentWithRevisionTurnedOn.get(doc.id)
        found_doc.num_1 += 1
        await found_doc.replace()

    doc._previous_revision_id = "wrong"
    doc.num_1 = 4
    with pytest.raises(RevisionIdWasChanged):
        await doc.replace()

    await doc.replace(ignore_revision=True)
    await doc.replace()


async def test_update():
    doc = DocumentWithRevisionTurnedOn(num_1=1, num_2=2)

    await doc.insert()

    await doc.update(Inc({DocumentWithRevisionTurnedOn.num_1: 1}))
    await doc.update(Inc({DocumentWithRevisionTurnedOn.num_1: 1}))

    for i in range(5):
        found_doc = await DocumentWithRevisionTurnedOn.get(doc.id)
        await found_doc.update(Inc({DocumentWithRevisionTurnedOn.num_1: 1}))

    doc._previous_revision_id = "wrong"
    with pytest.raises(RevisionIdWasChanged):
        await doc.update(Inc({DocumentWithRevisionTurnedOn.num_1: 1}))

    await doc.update(
        Inc({DocumentWithRevisionTurnedOn.num_1: 1}), ignore_revision=True
    )
    await doc.update(Inc({DocumentWithRevisionTurnedOn.num_1: 1}))


async def test_save_changes():
    doc = DocumentWithRevisionTurnedOn(num_1=1, num_2=2)
    await doc.insert()

    doc.num_1 = 2
    await doc.save_changes()

    doc.num_2 = 3
    await doc.save_changes()

    for i in range(5):
        found_doc = await DocumentWithRevisionTurnedOn.get(doc.id)
        found_doc.num_1 += 1
        await found_doc.save_changes()

    doc._previous_revision_id = "wrong"
    doc.num_1 = 4
    with pytest.raises(RevisionIdWasChanged):
        await doc.save_changes()

    await doc.save_changes(ignore_revision=True)
    await doc.save_changes()


async def test_save():
    doc = DocumentWithRevisionTurnedOn(num_1=1, num_2=2)

    doc.num_1 = 2
    await doc.save()

    doc.num_2 = 3
    await doc.save()

    for i in range(5):
        found_doc = await DocumentWithRevisionTurnedOn.get(doc.id)
        found_doc.num_1 += 1
        await found_doc.save()

    doc._previous_revision_id = "wrong"
    doc.num_1 = 4
    with pytest.raises(RevisionIdWasChanged):
        await doc.save()

    await doc.save(ignore_revision=True)
    await doc.save()


async def test_update_bulk_writer():
    doc = DocumentWithRevisionTurnedOn(num_1=1, num_2=2)
    await doc.save()

    doc.num_1 = 2
    async with BulkWriter() as bulk_writer:
        await doc.save(bulk_writer=bulk_writer)

    doc = await DocumentWithRevisionTurnedOn.get(doc.id)

    doc.num_2 = 3
    async with BulkWriter() as bulk_writer:
        await doc.save(bulk_writer=bulk_writer)

    doc = await DocumentWithRevisionTurnedOn.get(doc.id)

    for i in range(5):
        found_doc = await DocumentWithRevisionTurnedOn.get(doc.id)
        found_doc.num_1 += 1
        async with BulkWriter() as bulk_writer:
            await found_doc.save(bulk_writer=bulk_writer)

    doc._previous_revision_id = "wrong"
    doc.num_1 = 4
    with pytest.raises(BulkWriteError):
        async with BulkWriter() as bulk_writer:
            await doc.save(bulk_writer=bulk_writer)

    async with BulkWriter() as bulk_writer:
        await doc.save(bulk_writer=bulk_writer, ignore_revision=True)


async def test_empty_update():
    doc = DocumentWithRevisionTurnedOn(num_1=1, num_2=2)
    await doc.insert()

    # This fails with RevisionIdWasChanged
    await doc.update({"$set": {"num_1": 1}})


async def test_save_changes_when_there_were_no_changes():
    doc = DocumentWithRevisionTurnedOn(num_1=1, num_2=2)
    await doc.insert()
    revision = doc.revision_id
    old_revision = doc._previous_revision_id

    await doc.save_changes()
    assert doc.revision_id == revision
    assert doc._previous_revision_id == old_revision

    doc = await DocumentWithRevisionTurnedOn.get(doc.id)
    assert doc._previous_revision_id == old_revision



from beanie.odm.enums import SortDirection
from beanie.odm.operators.find.comparison import In, NotIn
from tests.odm.models import Sample


def test_nesting():
    assert Sample.id == "_id"

    q = Sample.find_many(Sample.integer == 1)
    assert q.get_filter_query() == {"integer": 1}
    assert Sample.integer == "integer"

    q = Sample.find_many(Sample.nested.integer == 1)
    assert q.get_filter_query() == {"nested.integer": 1}
    assert Sample.nested.integer == "nested.integer"

    q = Sample.find_many(Sample.union.s == "test")
    assert q.get_filter_query() == {"union.s": "test"}
    assert Sample.union.s == "union.s"

    q = Sample.find_many(Sample.nested.optional == None)  # noqa
    assert q.get_filter_query() == {"nested.optional": None}
    assert Sample.nested.optional == "nested.optional"

    q = Sample.find_many(Sample.nested.integer == 1).find_many(
        Sample.nested.union.s == "test"
    )
    assert q.get_filter_query() == {
        "$and": [{"nested.integer": 1}, {"nested.union.s": "test"}]
    }


def test_eq():
    q = Sample.find_many(Sample.integer == 1)
    assert q.get_filter_query() == {"integer": 1}


def test_gt():
    q = Sample.find_many(Sample.integer > 1)
    assert q.get_filter_query() == {"integer": {"$gt": 1}}


def test_gte():
    q = Sample.find_many(Sample.integer >= 1)
    assert q.get_filter_query() == {"integer": {"$gte": 1}}


def test_in():
    q = Sample.find_many(In(Sample.integer, [1, 2, 3, 4]))
    assert dict(q.get_filter_query()) == {"integer": {"$in": [1, 2, 3, 4]}}


def test_lt():
    q = Sample.find_many(Sample.integer < 1)
    assert q.get_filter_query() == {"integer": {"$lt": 1}}


def test_lte():
    q = Sample.find_many(Sample.integer <= 1)
    assert q.get_filter_query() == {"integer": {"$lte": 1}}


def test_ne():
    q = Sample.find_many(Sample.integer != 1)
    assert q.get_filter_query() == {"integer": {"$ne": 1}}


def test_nin():
    q = Sample.find_many(NotIn(Sample.integer, [1, 2, 3, 4]))
    assert dict(q.get_filter_query()) == {"integer": {"$nin": [1, 2, 3, 4]}}


def test_pos():
    q = +Sample.integer
    assert q == ("integer", SortDirection.ASCENDING)


def test_neg():
    q = -Sample.integer
    assert q == ("integer", SortDirection.DESCENDING)



import pytest
from pydantic import ValidationError

from tests.odm.models import DocumentWithValidationOnSave


async def test_validate_on_insert():
    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    doc.num_1 = "wrong_value"
    with pytest.raises(ValidationError):
        await doc.insert()


async def test_validate_on_replace():
    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    await doc.insert()
    doc.num_1 = "wrong_value"
    with pytest.raises(ValidationError):
        await doc.replace()


async def test_validate_on_save_changes():
    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    await doc.insert()
    doc.num_1 = "wrong_value"
    with pytest.raises(ValidationError):
        await doc.save_changes()


async def test_validate_on_save_action():
    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    await doc.insert()
    assert doc.num_2 == 3


async def test_validate_on_save_skip_action():
    doc = DocumentWithValidationOnSave(num_1=1, num_2=2)
    await doc.insert(skip_actions=["num_2_plus_1"])
    assert doc.num_2 == 2



import asyncio
import sys
from collections import OrderedDict
from enum import Enum
from typing import (
    Dict,
    Generic,
    TypeVar,
    Union,
    Type,
    List,
    Optional,
    Any,
    TYPE_CHECKING,
)

if sys.version_info >= (3, 8):
    from typing import get_args
else:
    from typing_extensions import get_args

from typing import OrderedDict as OrderedDictType

from bson import ObjectId, DBRef
from bson.errors import InvalidId
from pydantic import BaseModel

from pymongo import ASCENDING

from beanie.odm.enums import SortDirection
from beanie.odm.operators.find.comparison import (
    Eq,
    GT,
    GTE,
    LT,
    LTE,
    NE,
    In,
)
from beanie.odm.utils.parsing import parse_obj
from pymongo import IndexModel

from beanie.odm.registry import DocsRegistry

from beanie.odm.utils.pydantic import (
    IS_PYDANTIC_V2,
    get_model_fields,
    get_field_type,
    parse_object_as,
)

if IS_PYDANTIC_V2:
    from pydantic import (
        GetCoreSchemaHandler,
        GetJsonSchemaHandler,
        TypeAdapter,
    )
    from pydantic.json_schema import JsonSchemaValue
    from pydantic_core import core_schema, CoreSchema
    from pydantic_core.core_schema import (
        simple_ser_schema,
        ValidationInfo,
        str_schema,
    )
else:
    from pydantic.fields import ModelField  # type: ignore
    from pydantic.json import ENCODERS_BY_TYPE

if TYPE_CHECKING:
    from beanie.odm.documents import DocType


def Indexed(typ, index_type=ASCENDING, **kwargs):
    """
    Returns a subclass of `typ` with an extra attribute `_indexed` as a tuple:
    - Index 0: `index_type` such as `pymongo.ASCENDING`
    - Index 1: `kwargs` passed to `IndexModel`
    When instantiated the type of the result will actually be `typ`.
    """

    class NewType(typ):
        _indexed = (index_type, kwargs)

        def __new__(cls, *args, **kwargs):
            return typ.__new__(typ, *args, **kwargs)

        if IS_PYDANTIC_V2:

            @classmethod
            def __get_pydantic_core_schema__(
                cls, _source_type: Any, _handler: GetCoreSchemaHandler
            ) -> core_schema.CoreSchema:
                return core_schema.no_info_after_validator_function(
                    lambda v: v,
                    simple_ser_schema(typ.__name__),
                )

    NewType.__name__ = f"Indexed {typ.__name__}"
    return NewType


class PydanticObjectId(ObjectId):
    """
    Object Id field. Compatible with Pydantic.
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    if IS_PYDANTIC_V2:

        @classmethod
        def validate(cls, v, _: ValidationInfo):
            if isinstance(v, bytes):
                v = v.decode("utf-8")
            try:
                return PydanticObjectId(v)
            except InvalidId:
                raise ValueError("Id must be of type PydanticObjectId")

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Any, handler: GetCoreSchemaHandler
        ) -> CoreSchema:  # type: ignore
            return core_schema.json_or_python_schema(
                python_schema=core_schema.general_plain_validator_function(
                    cls.validate
                ),
                json_schema=str_schema(),
                serialization=core_schema.plain_serializer_function_ser_schema(
                    lambda instance: str(instance)
                ),
            )

        @classmethod
        def __get_pydantic_json_schema__(
            cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler  # type: ignore
        ) -> JsonSchemaValue:
            json_schema = handler(schema)
            json_schema.update(
                type="string",
                example="5eb7cf5a86d9755df3a6c593",
            )
            return json_schema

    else:

        @classmethod
        def validate(cls, v):
            if isinstance(v, bytes):
                v = v.decode("utf-8")
            try:
                return PydanticObjectId(v)
            except InvalidId:
                raise TypeError("Id must be of type PydanticObjectId")

        @classmethod
        def __modify_schema__(cls, field_schema):
            field_schema.update(
                type="string",
                example="5eb7cf5a86d9755df3a6c593",
            )


if not IS_PYDANTIC_V2:
    ENCODERS_BY_TYPE[
        PydanticObjectId
    ] = str  # it is a workaround to force pydantic make json schema for this field


class ExpressionField(str):
    def __getitem__(self, item):
        """
        Get sub field

        :param item: name of the subfield
        :return: ExpressionField
        """
        return ExpressionField(f"{self}.{item}")

    def __getattr__(self, item):
        """
        Get sub field

        :param item: name of the subfield
        :return: ExpressionField
        """
        return ExpressionField(f"{self}.{item}")

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if isinstance(other, ExpressionField):
            return super(ExpressionField, self).__eq__(other)
        return Eq(field=self, other=other)

    def __gt__(self, other):
        return GT(field=self, other=other)

    def __ge__(self, other):
        return GTE(field=self, other=other)

    def __lt__(self, other):
        return LT(field=self, other=other)

    def __le__(self, other):
        return LTE(field=self, other=other)

    def __ne__(self, other):
        return NE(field=self, other=other)

    def __pos__(self):
        return self, SortDirection.ASCENDING

    def __neg__(self):
        return self, SortDirection.DESCENDING

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self


class DeleteRules(str, Enum):
    DO_NOTHING = "DO_NOTHING"
    DELETE_LINKS = "DELETE_LINKS"


class WriteRules(str, Enum):
    DO_NOTHING = "DO_NOTHING"
    WRITE = "WRITE"


class LinkTypes(str, Enum):
    DIRECT = "DIRECT"
    OPTIONAL_DIRECT = "OPTIONAL_DIRECT"
    LIST = "LIST"
    OPTIONAL_LIST = "OPTIONAL_LIST"

    BACK_DIRECT = "BACK_DIRECT"
    BACK_LIST = "BACK_LIST"
    OPTIONAL_BACK_DIRECT = "OPTIONAL_BACK_DIRECT"
    OPTIONAL_BACK_LIST = "OPTIONAL_BACK_LIST"


class LinkInfo(BaseModel):
    field_name: str
    lookup_field_name: str
    document_class: Type[BaseModel]  # Document class
    link_type: LinkTypes
    nested_links: Optional[Dict] = None


T = TypeVar("T")


class Link(Generic[T]):
    def __init__(self, ref: DBRef, document_class: Type[T]):
        self.ref = ref
        self.document_class = document_class

    async def fetch(self, fetch_links: bool = False) -> Union[T, "Link"]:
        result = await self.document_class.get(  # type: ignore
            self.ref.id, with_children=True, fetch_links=fetch_links
        )
        return result or self

    @classmethod
    async def fetch_one(cls, link: "Link"):
        return await link.fetch()

    @classmethod
    async def fetch_list(
        cls, links: List[Union["Link", "DocType"]], fetch_links: bool = False
    ):
        """
        Fetch list that contains links and documents
        :param links:
        :param fetch_links:
        :return:
        """
        data = Link.repack_links(links)  # type: ignore
        ids_to_fetch = []
        document_class = None
        for doc_id, link in data.items():
            if isinstance(link, Link):
                if document_class is None:
                    document_class = link.document_class
                else:
                    if document_class != link.document_class:
                        raise ValueError(
                            "All the links must have the same model class"
                        )
                ids_to_fetch.append(link.ref.id)

        fetched_models = await document_class.find(  # type: ignore
            In("_id", ids_to_fetch),
            with_children=True,
            fetch_links=fetch_links,
        ).to_list()

        for model in fetched_models:
            data[model.id] = model

        return list(data.values())

    @staticmethod
    def repack_links(
        links: List[Union["Link", "DocType"]]
    ) -> OrderedDictType[Any, Any]:
        result = OrderedDict()
        for link in links:
            if isinstance(link, Link):
                result[link.ref.id] = link
            else:
                result[link.id] = link
        return result

    @classmethod
    async def fetch_many(cls, links: List["Link"]):
        coros = []
        for link in links:
            coros.append(link.fetch())
        return await asyncio.gather(*coros)

    if IS_PYDANTIC_V2:

        @staticmethod
        def serialize(value: Union["Link", BaseModel]):
            if isinstance(value, Link):
                return value.to_dict()
            return value.model_dump()

        @classmethod
        def build_validation(cls, handler, source_type):
            def validate(v: Union[DBRef, T], validation_info: ValidationInfo):
                document_class = DocsRegistry.evaluate_fr(get_args(source_type)[0])  # type: ignore  # noqa: F821

                if isinstance(v, DBRef):
                    return cls(ref=v, document_class=document_class)
                if isinstance(v, Link):
                    return v
                if isinstance(v, dict) or isinstance(v, BaseModel):
                    return parse_obj(document_class, v)
                new_id = TypeAdapter(
                    document_class.model_fields["id"].annotation
                ).validate_python(v)
                ref = DBRef(
                    collection=document_class.get_collection_name(), id=new_id
                )
                return cls(ref=ref, document_class=document_class)

            return validate

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Any, handler: GetCoreSchemaHandler
        ) -> CoreSchema:  # type: ignore
            return core_schema.json_or_python_schema(
                python_schema=core_schema.general_plain_validator_function(
                    cls.build_validation(handler, source_type)
                ),
                json_schema=core_schema.typed_dict_schema(
                    {
                        "id": core_schema.typed_dict_field(
                            core_schema.str_schema()
                        ),
                        "collection": core_schema.typed_dict_field(
                            core_schema.str_schema()
                        ),
                    }
                ),
                serialization=core_schema.plain_serializer_function_ser_schema(  # type: ignore
                    lambda instance: cls.serialize(instance)  # type: ignore
                ),
            )
            return core_schema.general_plain_validator_function(
                cls.build_validation(handler, source_type)
            )

    else:

        @classmethod
        def __get_validators__(cls):
            yield cls.validate

        @classmethod
        def validate(cls, v: Union[DBRef, T], field: ModelField):
            document_class = field.sub_fields[0].type_  # type: ignore
            if isinstance(v, DBRef):
                return cls(ref=v, document_class=document_class)
            if isinstance(v, Link):
                return v
            if isinstance(v, dict) or isinstance(v, BaseModel):
                return parse_obj(document_class, v)
            new_id = parse_object_as(
                get_field_type(get_model_fields(document_class)["id"]), v
            )
            ref = DBRef(
                collection=document_class.get_collection_name(), id=new_id
            )
            return cls(ref=ref, document_class=document_class)

    def to_ref(self):
        return self.ref

    def to_dict(self):
        return {"id": str(self.ref.id), "collection": self.ref.collection}


if not IS_PYDANTIC_V2:
    ENCODERS_BY_TYPE[Link] = lambda o: o.to_dict()


class BackLink(Generic[T]):
    """Back reference to a document"""

    def __init__(self, document_class: Type[T]):
        self.document_class = document_class

    if IS_PYDANTIC_V2:

        @classmethod
        def build_validation(cls, handler, source_type):
            def validate(v: Union[DBRef, T], field):
                document_class = DocsRegistry.evaluate_fr(get_args(source_type)[0])  # type: ignore  # noqa: F821
                if isinstance(v, dict) or isinstance(v, BaseModel):
                    return parse_obj(document_class, v)
                return cls(document_class=document_class)

            return validate

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Any, handler: GetCoreSchemaHandler
        ) -> CoreSchema:  # type: ignore
            return core_schema.general_plain_validator_function(
                cls.build_validation(handler, source_type)
            )

    else:

        @classmethod
        def __get_validators__(cls):
            yield cls.validate

        @classmethod
        def validate(cls, v: Union[DBRef, T], field: ModelField):
            document_class = field.sub_fields[0].type_  # type: ignore
            if isinstance(v, dict) or isinstance(v, BaseModel):
                return parse_obj(document_class, v)
            return cls(document_class=document_class)

    def to_dict(self):
        return {"collection": self.document_class.get_collection_name()}


if not IS_PYDANTIC_V2:
    ENCODERS_BY_TYPE[BackLink] = lambda o: o.to_dict()


class IndexModelField:
    def __init__(self, index: IndexModel):
        self.index = index
        self.name = index.document["name"]

        self.fields = tuple(sorted(self.index.document["key"]))
        self.options = tuple(
            sorted(
                (k, v)
                for k, v in self.index.document.items()
                if k not in ["key", "v"]
            )
        )

    def __eq__(self, other):
        return self.fields == other.fields and self.options == other.options

    def __repr__(self):
        return f"IndexModelField({self.name}, {self.fields}, {self.options})"

    @staticmethod
    def list_difference(
        left: List["IndexModelField"], right: List["IndexModelField"]
    ):
        result = []
        for index in left:
            if index not in right:
                result.append(index)
        return result

    @staticmethod
    def list_to_index_model(left: List["IndexModelField"]):
        return [index.index for index in left]

    @classmethod
    def from_motor_index_information(cls, index_info: dict):
        result = []
        for name, details in index_info.items():
            fields = details["key"]
            if ("_id", 1) in fields:
                continue

            options = {k: v for k, v in details.items() if k != "key"}
            index_model = IndexModelField(
                IndexModel(fields, name=name, **options)
            )
            result.append(index_model)
        return result

    def same_fields(self, other: "IndexModelField"):
        return self.fields == other.fields

    @staticmethod
    def find_index_with_the_same_fields(
        indexes: List["IndexModelField"], index: "IndexModelField"
    ):
        for i in indexes:
            if i.same_fields(index):
                return i
        return None

    @staticmethod
    def merge_indexes(
        left: List["IndexModelField"], right: List["IndexModelField"]
    ):
        left_dict = {index.fields: index for index in left}
        right_dict = {index.fields: index for index in right}
        left_dict.update(right_dict)
        return list(left_dict.values())

    if IS_PYDANTIC_V2:

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Any, handler: GetCoreSchemaHandler
        ) -> CoreSchema:  # type: ignore
            def validate(v, _):
                if isinstance(v, IndexModel):
                    return IndexModelField(v)
                else:
                    return IndexModelField(IndexModel(v))

            return core_schema.general_plain_validator_function(validate)

    else:

        @classmethod
        def __get_validators__(cls):
            yield cls.validate

        @classmethod
        def validate(cls, v):
            if isinstance(v, IndexModel):
                return IndexModelField(v)
            else:
                return IndexModelField(IndexModel(v))



from beanie.odm.operators.update.general import (
    Set,
    CurrentDate,
    Inc,
    Unset,
    SetOnInsert,
    Rename,
    Mul,
    Max,
    Min,
)
from tests.odm.models import Sample


def test_set():
    q = Set({Sample.integer: 2})
    assert q == {"$set": {"integer": 2}}


def test_current_date():
    q = CurrentDate({Sample.integer: 2})
    assert q == {"$currentDate": {"integer": 2}}


def test_inc():
    q = Inc({Sample.integer: 2})
    assert q == {"$inc": {"integer": 2}}


def test_min():
    q = Min({Sample.integer: 2})
    assert q == {"$min": {"integer": 2}}


def test_max():
    q = Max({Sample.integer: 2})
    assert q == {"$max": {"integer": 2}}


def test_mul():
    q = Mul({Sample.integer: 2})
    assert q == {"$mul": {"integer": 2}}


def test_rename():
    q = Rename({Sample.integer: 2})
    assert q == {"$rename": {"integer": 2}}


def test_set_on_insert():
    q = SetOnInsert({Sample.integer: 2})
    assert q == {"$setOnInsert": {"integer": 2}}


def test_unset():
    q = Unset({Sample.integer: 2})
    assert q == {"$unset": {"integer": 2}}



from beanie.odm.operators.update.general import Set
from tests.odm.models import (
    DocumentTestModel,
    DocumentWithTimeStampToTestConsistency,
)


class TestResponseOfTheChangingOperations:
    async def test_insert(self, document_not_inserted):
        result = await document_not_inserted.insert()
        assert isinstance(result, DocumentTestModel)

    async def test_update(self, document):
        result = await document.update(Set({"test_int": 43}))
        assert isinstance(result, DocumentTestModel)

    async def test_save(self, document, document_not_inserted):
        document.test_int = 43
        result = await document.save()
        assert isinstance(result, DocumentTestModel)

        document_not_inserted.test_int = 43
        result = await document_not_inserted.save()
        assert isinstance(result, DocumentTestModel)

    async def test_save_changes(self, document):
        document.test_int = 43
        result = await document.save_changes()
        assert isinstance(result, DocumentTestModel)

    async def test_replace(self, document):
        result = await document.replace()
        assert isinstance(result, DocumentTestModel)

    async def test_set(self, document):
        result = await document.set({"test_int": 43})
        assert isinstance(result, DocumentTestModel)

    async def test_inc(self, document):
        result = await document.inc({"test_int": 1})
        assert isinstance(result, DocumentTestModel)

    async def test_current_date(self):
        document = DocumentWithTimeStampToTestConsistency()
        await document.insert()
        result = await document.current_date({"ts": True})
        assert isinstance(result, DocumentWithTimeStampToTestConsistency)



from tests.odm.models import DocumentTestModel


async def test_delete_one(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    await DocumentTestModel.find_one({"test_str": "uno"}).delete()
    documents = await DocumentTestModel.find_all().to_list()
    assert len(documents) == 6


async def test_delete_one_not_found(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    await DocumentTestModel.find_one({"test_str": "wrong"}).delete()
    documents = await DocumentTestModel.find_all().to_list()
    assert len(documents) == 7


async def test_delete_many(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    await DocumentTestModel.find_many({"test_str": "uno"}).delete()
    documents = await DocumentTestModel.find_all().to_list()
    assert len(documents) == 3


async def test_delete_many_not_found(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    await DocumentTestModel.find_many({"test_str": "wrong"}).delete()
    documents = await DocumentTestModel.find_all().to_list()
    assert len(documents) == 7


async def test_delete_all(documents):
    await documents(4, "uno")
    await documents(2, "dos")
    await documents(1, "cuatro")
    await DocumentTestModel.delete_all()
    documents = await DocumentTestModel.find_all().to_list()
    assert len(documents) == 0


async def test_delete(document):
    doc_id = document.id
    await document.delete()
    new_document = await DocumentTestModel.get(doc_id)
    assert new_document is None



from pydantic.main import BaseModel

from beanie import Document, iterative_migration


class Tag(BaseModel):
    color: str
    name: str


class OldNote(Document):
    title: str
    tag_name: str
    tag_color: str

    class Settings:
        name = "notes"


class Note(Document):
    title: str
    tag: Tag

    class Settings:
        name = "notes"


class Forward:
    @iterative_migration()
    async def pack(self, input_document: OldNote, output_document: Note):
        output_document.tag = Tag(
            name=input_document.tag_name, color=input_document.tag_color
        )


class Backward:
    @iterative_migration()
    async def unpack(self, input_document: Note, output_document: OldNote):
        output_document.tag_name = input_document.tag.name
        output_document.tag_color = input_document.tag.color



import sys

if sys.version_info >= (3, 8):
    from typing import get_args, get_origin
else:
    from typing_extensions import get_args, get_origin

from typing import Optional, Union, Type, Any
import inspect


def extract_id_class(annotation) -> Type[Any]:
    if inspect.isclass(annotation):
        return annotation

    elif get_origin(annotation) is Union:
        args = get_args(annotation)
        for arg in args:
            if inspect.isclass(arg) and arg is not type(None):
                return arg
        raise ValueError("Unknown annotation: {}".format(annotation))

    elif get_origin(annotation) is Optional:
        arg = get_args(annotation)[0]
        if inspect.isclass(arg):
            return arg
        else:
            raise ValueError("Unknown annotation: {}".format(annotation))
    else:
        raise ValueError("Unknown annotation: {}".format(annotation))



# check python version
import sys

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

from decimal import Decimal as NativeDecimal
from typing import Any, Callable

from bson import Decimal128
from pydantic import GetJsonSchemaHandler
from pydantic.fields import FieldInfo
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema


class DecimalCustomAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],  # type: ignore
    ) -> core_schema.CoreSchema:  # type: ignore
        def validate(value, _: FieldInfo) -> NativeDecimal:
            if isinstance(value, Decimal128):
                return value.to_decimal()
            return value

        python_schema = core_schema.general_plain_validator_function(validate)  # type: ignore

        return core_schema.json_or_python_schema(
            json_schema=core_schema.float_schema(),
            python_schema=python_schema,
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _core_schema: core_schema.CoreSchema,  # type: ignore
        handler: GetJsonSchemaHandler,
    ) -> JsonSchemaValue:
        return handler(core_schema.float_schema())


DecimalAnnotation = Annotated[NativeDecimal, DecimalCustomAnnotation]



import asyncio

import pytest

from beanie.odm.operators.update.general import Set, Max
from tests.odm.models import Sample


async def test_update_query():
    q = (
        Sample.find_many(Sample.integer == 1)
        .update(Set({Sample.integer: 10}))
        .update_query
    )
    assert q == {"$set": {"integer": 10}}

    q = (
        Sample.find_many(Sample.integer == 1)
        .update(Max({Sample.integer: 10}), Set({Sample.optional: None}))
        .update_query
    )
    assert q == {"$max": {"integer": 10}, "$set": {"optional": None}}

    q = (
        Sample.find_many(Sample.integer == 1)
        .update(Set({Sample.integer: 10}), Set({Sample.optional: None}))
        .update_query
    )
    assert q == {"$set": {"optional": None}}

    q = (
        Sample.find_many(Sample.integer == 1)
        .update(Max({Sample.integer: 10}))
        .update(Set({Sample.optional: None}))
        .update_query
    )
    assert q == {"$max": {"integer": 10}, "$set": {"optional": None}}

    q = (
        Sample.find_many(Sample.integer == 1)
        .update(Set({Sample.integer: 10}))
        .update(Set({Sample.optional: None}))
        .update_query
    )
    assert q == {"$set": {"optional": None}}

    with pytest.raises(TypeError):
        Sample.find_many(Sample.integer == 1).update(40).update_query


async def test_update_many(preset_documents):
    await Sample.find_many(Sample.increment > 4).find_many(
        Sample.nested.optional == None
    ).update(
        Set({Sample.increment: 100})
    )  # noqa
    result = await Sample.find_many(Sample.increment == 100).to_list()
    assert len(result) == 3
    for sample in result:
        assert sample.increment == 100


async def test_update_many_linked_method(preset_documents):
    await Sample.find_many(Sample.increment > 4).find_many(
        Sample.nested.optional == None
    ).update_many(
        Set({Sample.increment: 100})
    )  # noqa
    result = await Sample.find_many(Sample.increment == 100).to_list()
    assert len(result) == 3
    for sample in result:
        assert sample.increment == 100


async def test_update_all(preset_documents):
    await Sample.update_all(Set({Sample.integer: 100}))
    result = await Sample.find_all().to_list()
    for sample in result:
        assert sample.integer == 100

    await Sample.find_all().update(Set({Sample.integer: 101}))
    result = await Sample.find_all().to_list()
    for sample in result:
        assert sample.integer == 101


async def test_update_one(preset_documents):
    await Sample.find_one(Sample.integer == 1).update(
        Set({Sample.integer: 100})
    )
    result = await Sample.find_many(Sample.integer == 100).to_list()
    assert len(result) == 1
    assert result[0].integer == 100

    await Sample.find_one(Sample.integer == 1).update_one(
        Set({Sample.integer: 101})
    )
    result = await Sample.find_many(Sample.integer == 101).to_list()
    assert len(result) == 1
    assert result[0].integer == 101


async def test_update_self(preset_documents):
    sample = await Sample.find_one(Sample.integer == 1)
    await sample.update(Set({Sample.integer: 100}))
    assert sample.integer == 100

    result = await Sample.find_many(Sample.integer == 100).to_list()
    assert len(result) == 1
    assert result[0].integer == 100


async def test_update_many_with_session(preset_documents, session):
    q = (
        Sample.find_many(Sample.increment > 4)
        .find_many(Sample.nested.optional == None)
        .update(Set({Sample.increment: 100}))
        .set_session(session=session)
    )
    assert q.session == session

    q = (
        Sample.find_many(Sample.increment > 4)
        .find_many(Sample.nested.optional == None)
        .update(Set({Sample.increment: 100}), session=session)
    )
    assert q.session == session

    q = (
        Sample.find_many(Sample.increment > 4)
        .find_many(Sample.nested.optional == None, session=session)
        .update(Set({Sample.increment: 100}))
    )
    assert q.session == session

    await q  # noqa
    result = await Sample.find_many(Sample.increment == 100).to_list()
    assert len(result) == 3
    for sample in result:
        assert sample.increment == 100


async def test_update_many_upsert_with_insert(
    preset_documents, sample_doc_not_saved
):
    await Sample.find_many(Sample.integer > 100000).upsert(
        Set({Sample.integer: 100}), on_insert=sample_doc_not_saved
    )
    await asyncio.sleep(2)
    new_docs = await Sample.find_many(
        Sample.string == sample_doc_not_saved.string
    ).to_list()
    assert len(new_docs) == 1
    doc = new_docs[0]
    assert doc.integer == sample_doc_not_saved.integer


async def test_update_many_upsert_without_insert(
    preset_documents, sample_doc_not_saved
):
    await Sample.find_many(Sample.integer > 1).upsert(
        Set({Sample.integer: 100}), on_insert=sample_doc_not_saved
    )
    await asyncio.sleep(2)
    new_docs = await Sample.find_many(
        Sample.string == sample_doc_not_saved.string
    ).to_list()
    assert len(new_docs) == 0


async def test_update_one_upsert_with_insert(
    preset_documents, sample_doc_not_saved
):
    await Sample.find_one(Sample.integer > 100000).upsert(
        Set({Sample.integer: 100}), on_insert=sample_doc_not_saved
    )
    new_docs = await Sample.find_many(
        Sample.string == sample_doc_not_saved.string
    ).to_list()
    assert len(new_docs) == 1
    doc = new_docs[0]
    assert doc.integer == sample_doc_not_saved.integer


async def test_update_one_upsert_without_insert(
    preset_documents, sample_doc_not_saved
):
    await Sample.find_one(Sample.integer > 1).upsert(
        Set({Sample.integer: 100}), on_insert=sample_doc_not_saved
    )
    new_docs = await Sample.find_many(
        Sample.string == sample_doc_not_saved.string
    ).to_list()
    assert len(new_docs) == 0


async def test_update_pymongo_kwargs(preset_documents):
    with pytest.raises(TypeError):
        await Sample.find_many(Sample.increment > 4).update(
            Set({Sample.increment: 100}), wrong="integer_1"
        )

    await Sample.find_many(Sample.increment > 4).update(
        Set({Sample.increment: 100}), hint="integer_1"
    )

    await Sample.find_one(Sample.increment > 4).update(
        Set({Sample.increment: 100}), hint="integer_1"
    )


def test_clone():
    q = Sample.find_many(Sample.integer == 1).update(Set({Sample.integer: 10}))
    new_q = q.clone()
    new_q.update(Max({Sample.integer: 10}))
    assert q.update_query == {"$set": {"integer": 10}}
    assert new_q.update_query == {
        "$max": {"integer": 10},
        "$set": {"integer": 10},
    }



from typing import List, Optional

from tests.typing.models import Test, ProjectionTest


async def find_many() -> List[Test]:
    return await Test.find().to_list()


async def find_many_with_projection() -> List[ProjectionTest]:
    return await Test.find().project(projection_model=ProjectionTest).to_list()


async def find_many_generator() -> List[Test]:
    docs: List[Test] = []
    async for doc in Test.find():
        docs.append(doc)
    return docs


async def find_many_generator_with_projection() -> List[ProjectionTest]:
    docs: List[ProjectionTest] = []
    async for doc in Test.find().project(projection_model=ProjectionTest):
        docs.append(doc)
    return docs


async def find_one() -> Optional[Test]:
    return await Test.find_one()


async def find_one_with_projection() -> Optional[ProjectionTest]:
    return await Test.find_one().project(projection_model=ProjectionTest)



from tests.odm.views import TestView, TestViewWithLink


class TestViews:
    async def test_simple(self, documents):
        await documents(number=15)
        results = await TestView.all().to_list()
        assert len(results) == 6

    async def test_aggregate(self, documents):
        await documents(number=15)
        results = await TestView.aggregate(
            [
                {"$set": {"test_field": 1}},
                {"$match": {"$expr": {"$lt": ["$number", 12]}}},
            ]
        ).to_list()
        assert len(results) == 3
        assert results[0]["test_field"] == 1

    async def test_link(self, documents_with_links):
        await documents_with_links()
        results = await TestViewWithLink.all().to_list()
        for document in results:
            await document.fetch_all_links()

        for i, document in enumerate(results):
            assert document.link.test_int == i



from beanie.odm.operators.find.logical import And, Not, Nor, Or
from tests.odm.models import Sample


async def test_and():
    q = And(Sample.integer == 1)
    assert q == {"integer": 1}

    q = And(Sample.integer == 1, Sample.nested.integer > 3)
    assert q == {"$and": [{"integer": 1}, {"nested.integer": {"$gt": 3}}]}


async def test_not():
    q = Not(Sample.integer == 1)
    assert q == {"$not": {"integer": 1}}


async def test_nor():
    q = Nor(Sample.integer == 1)
    assert q == {"$nor": [{"integer": 1}]}

    q = Nor(Sample.integer == 1, Sample.nested.integer > 3)
    assert q == {"$nor": [{"integer": 1}, {"nested.integer": {"$gt": 3}}]}


async def test_or():
    q = Or(Sample.integer == 1)
    assert q == {"integer": 1}

    q = Or(Sample.integer == 1, Sample.nested.integer > 3)
    assert q == {"$or": [{"integer": 1}, {"nested.integer": {"$gt": 3}}]}
